#include <iostream>
#include <fstream>
#include <vector>

using namespace std;

#define uint unsigned int

///////////////////////////////////////////////////////////////
// Seq Klasse

class Seq{
    public:
        Seq();
        ~Seq();
        void setLength(uint len);
        void setFilename(string s){filename = s;}

        char* getBuf(){return buf;}
        uint getLength(){return length;}
        string getFilename(){return filename;}
        int indexOf(string s, int from);
        bool contains(string s);
        int total(string s);
        void removeStr(string s);

    private:
        char *buf;
        uint length;
        string filename;
};

Seq::Seq(){
    buf = 0;
    length = 0;
}

Seq::~Seq(){
    delete[] buf;
}

void Seq::setLength(unsigned int len){
    buf = new char[len];
    length = len;
}

int Seq::indexOf(string str, int from){
    for(uint i=from; i<length; i++){
        bool ok = true;
        for(uint j=0; j<str.length(); j++){
            if(buf[i+j] != str.at(j)){
                ok = false;
                break;
            }
        }
        if(ok)
            return i;
    }
    return -1;
}

bool Seq::contains(string str){
    int pos = indexOf(str, 0);
    if(pos == -1)
        return false;
    return true;
}

int Seq::total(string str){
    int pos = 0;
    int count = 0;
    while(true){
        pos = indexOf(str, pos);
        if(pos > -1){
            pos += str.length();
            count++;
        }
        else
            break;
    }
    return count;
}

void Seq::removeStr(string str){
    int pos = 0;
    while(true){
        pos = indexOf(str, pos);
        if(pos > -1){
            for(uint i=0; i<str.length(); i++){
                buf[pos+i] = 'x';
            }
            pos++;
        }
        else
            break;
    }
}

///////////////////////////////////////////////////////////////
// Funktionen

struct block{
    string s;
    int total;
};

void insert(vector<block> &v, block b){
    vector<block>::iterator it;
    it = v.begin();
    for(uint i=0; i<v.size(); i++){
        if(v.at(i).total < b.total){
            v.insert(it+i, b);
            return;
        }
    }
    v.push_back(b);
}

bool isContainsStr(vector<string> &v, string s){
    for(uint i=0; i<v.size(); i++){
        if(v.at(i) == s)
            return true;
    }
    return false;
}

bool isContainsInt(vector<int> &v, int s){
    for(uint i=0; i<v.size(); i++){
        if(v.at(i) == s)
            return true;
    }
    return false;
}

bool isContainsAllSeq(Seq seq[], int seqCount, string s){
    if(seqCount < 1)
        return false;
    int t = seq[0].total(s);
    for(int i=1; i<seqCount; i++){
        if(seq[i].total(s) != t)
            return false;
    }
    return true;
}

bool isContainsAllUSeq(Seq seq[], vector<int> &v, string s){
    if(v.size() < 1)
        return false;
    int t = seq[v.at(0)].total(s);
    for(uint i=1; i<v.size(); i++){
        if(seq[v.at(i)].total(s) != t)
            return false;
    }
    return true;
}

void selectBlocks(Seq *seq, vector<string> &blocks){
    string s;
    for(uint i=0; i<seq->getLength(); i++){
        char c = seq->getBuf()[i];
        if(c != 'x'){
            s.append(&c, 1);
        }
        else{
            if(s.length() > 0){
                if(!isContainsStr(blocks, s))
                    blocks.push_back(s);
                s.clear();
            }
        }
    }
}

///////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
    if(argc < 2){
        cout << "Usage: " << argv[0] << " <filename> ..." << endl;
        return 1;
    }

    int seqCount = argc - 1;
    Seq seq[seqCount];
    bool isError = false;

    /////////////////////////////////////////////////////////
    // Dateien lesen und in Buffer schreiben

    for(int i=0; i<seqCount; i++){
        ifstream is (argv[i+1], ifstream::binary);
        if (is) {
            is.seekg (0, is.end);
            seq[i].setLength(is.tellg());
            is.seekg (0, is.beg);

            is.read (seq[i].getBuf(), seq[i].getLength());

            is.close();

            seq[i].setFilename(argv[i+1]);
        }
        else
            isError = true;
    }
    if(isError){
        cout << "Error load Streams" << endl;
        return 1;
    }

    /////////////////////////////////////////////////////////
    // Buchstaben suchen und ausgeben

    cout << "-------------------------------------------" << endl;
    cout << "Buchstaben der einzelnen Sequenzen" << endl;

    uint len = 1;
    vector<int> uSeq;
    for(int k=0; k<seqCount; k++){
        cout << "-------------------------------------------" << endl;
        cout << seq[k].getFilename() << endl;
        cout << "Filesize: " << seq[k].getLength() << endl;
        vector<string> strings;
        for(uint i=0; i<seq[k].getLength(); i++){
            string s(&seq[k].getBuf()[i], len);
            if(isContainsStr(strings, s))
                continue;
            else
                strings.push_back(s);
            cout << s << " " << seq[k].total(s) << endl;

            // Sequenzen mit 'U' merken
            if(s == "U")
                uSeq.push_back(k);
        }
    }

    /////////////////////////////////////////////////////////
    // Ausgabe der Sequenzen mit und ohne 'U'

    cout << "-------------------------------------------" << endl;
    cout << "mit U" << endl;
    for(uint k=0; k<uSeq.size(); k++){
        cout << seq[uSeq.at(k)].getFilename() << endl;
    }
    cout << "Total: " << uSeq.size() << endl;

    cout << "-------------------------------------------" << endl;
    cout << "ohne U" << endl;
    vector<int> noUSeq;
    for(int k=0; k<seqCount; k++){
        if(!isContainsInt(uSeq, k)){
            noUSeq.push_back(k);
        }
    }
    for(uint k=0; k<noUSeq.size(); k++){
        cout << seq[noUSeq.at(k)].getFilename() << endl;
    }
    cout << "Total: " << noUSeq.size() << endl;

    if(uSeq.size() < 1){
        cout << "Error: nicht genuegend 'U'-Sequenzen" << endl;
        return 1;
    }

    if(noUSeq.size() < 1){
        cout << "Error: nicht genuegend nicht 'U'-Sequenzen" << endl;
        return 1;
    }

    /////////////////////////////////////////////////////////
    // Bausteine suchen und ausgeben

    cout << "-------------------------------------------" << endl;
    cout << "Bausteine von allen Sequenzen" << endl;
    cout << "extrahiert aus einer 'U'-Sequenz" << endl;

    // erste 'U'-Sequenz zum extrahieren verwenden
    Seq *seqU1 = &seq[uSeq.at(0)];
    cout << "Filename: " << seqU1->getFilename() << endl;
    cout << "-------------------------------------------" << endl;

    vector<string> blocks;
    len = 2;
    for(uint i=0; i+len<seqU1->getLength(); i+=len){
        string s;
        bool isAll = false;
        len = seqU1->getLength() - i -1;
        if(len > 20)
            len = 20;
        for(; len>2 && i+len<seqU1->getLength(); len--){
            s.clear();
            s.append(&seqU1->getBuf()[i], len);

            isAll = false;
            if(!isContainsStr(blocks, s)){
                // kommt dieser Baustein in allen 'U'-Sequenzen gleich oft vor?
                size_t found = s.find("U");
                if(found != string::npos){
                    if(isContainsAllUSeq(seq, uSeq, s)){
                        isAll = true;
                    }
                }
                // kommt dieser Baustein in allen Sequenzen gleich oft vor?
                else if(isContainsAllSeq(seq, seqCount, s)){
                    isAll = true;
                }
                if(isAll){
                    blocks.push_back(s);
                    cout << s << endl;
                    break;
                }
            }
            else{
                isAll = true;
                break;
            }
        }
        if(!isAll)
            len = 1;
    }
    cout << "-------------------------------------------" << endl;
    cout << "Bausteine total: " << blocks.size() << endl;

    /////////////////////////////////////////////////////////
    // Anzahl der Bausteine fuer jede Sequenz ausgeben

    cout << "-------------------------------------------" << endl;
    cout << "Bausteine der einzelnen Sequenzen" << endl;

    for(int k=0; k<seqCount; k++){
        cout << "-------------------------------------------" << endl;
        cout << seq[k].getFilename() << endl;
        cout << "Filesize: " << seq[k].getLength() << endl;
        int total = 0;
        for(uint i=0; i<blocks.size(); i++){
            int count = seq[k].total(blocks.at(i));
            cout << i+1 << " " << blocks.at(i) << " " << count << endl;
            total += count;
        }
        cout << "Total: " << total << endl;
    }

    /////////////////////////////////////////////////////////
    // Loeschen der Bausteine aus den Sequenzen

    cout << "-------------------------------------------" << endl;
    cout << "Buffer von den gefundenen Bausteinen bereinigen" << endl;
    cout << "-------------------------------------------" << endl;

    for(int k=0; k<seqCount; k++){
        for(uint i=0; i<blocks.size(); i++){
            seq[k].removeStr(blocks.at(i));
        }
        cout << k+1 << " ";
        cout.write(seq[k].getBuf(), seq[k].getLength());
        cout << endl;
    }

    /////////////////////////////////////////////////////////
    // restliche Bausteine suchen und ausgeben

    // erste 'U'-Sequenz zum extrahieren verwenden
    cout << "-------------------------------------------" << endl;
    cout << "restliche Bausteine einer 'U'-Sequenz" << endl;
    cout << "Filename: " << seqU1->getFilename() << endl;
    cout << "-------------------------------------------" << endl;

    blocks.clear();
    selectBlocks(seqU1, blocks);

    vector<block> blocksSort;
    for(uint i=0; i<blocks.size(); i++){
        block b;
        b.s = blocks.at(i);
        b.total = seqU1->total(blocks.at(i));
        insert(blocksSort, b);
    }
    for(uint i=0; i<blocksSort.size(); i++){
        cout << blocksSort.at(i).s << " " << blocksSort.at(i).total << endl;
    }

    cout << "-------------------------------------------" << endl;
    cout << "nun sind zwei Bausteine sehr gut zu erkennen" << endl;
    cout << "-------------------------------------------" << endl;

    // es fehlen nur noch 2 Bausteine, die 2 die am häufigsten vorkommen, sind es
    string b1;
    string b2;
    if(blocksSort.size() > 1){
        b1 = blocksSort.at(0).s;
        b2 = blocksSort.at(1).s;
    }
    else{
        cout << "Error: Es wurden keine 2 Bausteine gefunden!" << endl;
        return 1;
    }

    for(int k=0; k<seqCount; k++){
        cout << seq[k].getFilename() << endl;
        int t1 = seq[k].total(b1);
        int t2 = seq[k].total(b2);
        cout << b1 << " " << t1 << endl;
        cout << b2 << " " << t2 << endl;
        cout << "Summe: " << t1 + t2 << endl;
        cout << "-------------------------------------------" << endl;
    }

    /////////////////////////////////////////////////////////
    // Loeschen der Bausteine aus den Sequenzen

    cout << "Buffer von den gefundenen Bausteinen bereinigen" << endl;
    cout << "-------------------------------------------" << endl;

    for(int k=0; k<seqCount; k++){
        seq[k].removeStr(b1);
        seq[k].removeStr(b2);
        cout << k+1 << " ";
        cout.write(seq[k].getBuf(), seq[k].getLength());
        cout << endl;
    }

    /////////////////////////////////////////////////////////
    // das andere Geschlecht bleibt uebrig

    // erste nicht 'U'-Sequenz zum extrahieren verwenden
    Seq *seqNoU1 = &seq[noUSeq.at(0)];
    cout << "-------------------------------------------" << endl;
    cout << "das andere Geschlecht bleibt uebrig" << endl;
    cout << "extrahiert aus einer nicht 'U'-Sequenz" << endl;
    cout << "Filename: " << seqNoU1->getFilename() << endl;
    cout << "-------------------------------------------" << endl;

    blocks.clear();
    selectBlocks(seqNoU1, blocks);
    for(uint i=0; i<blocks.size(); i++){
        cout << blocks.at(i) << " " << seqNoU1->total(blocks.at(i)) << endl;
    }
    cout << "-------------------------------------------" << endl;
    cout << "---------------- ENDE ---------------------" << endl;

    return 0;
}

